#
# Autogenerated by Thrift Compiler (0.9.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException

from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
  from thrift.protocol import fastbinary
except:
  fastbinary = None


class ErrorCode:
  UNKNOWN = 0
  VERSION_CONFLICT = 1
  NOT_FOUND = 2
  INPUT_ERROR = 3

  _VALUES_TO_NAMES = {
    0: "UNKNOWN",
    1: "VERSION_CONFLICT",
    2: "NOT_FOUND",
    3: "INPUT_ERROR",
  }

  _NAMES_TO_VALUES = {
    "UNKNOWN": 0,
    "VERSION_CONFLICT": 1,
    "NOT_FOUND": 2,
    "INPUT_ERROR": 3,
  }


class Token:
  """
  Attributes:
   - version
   - name
   - owner
   - expirationTime
   - priority
   - data
  """

  thrift_spec = (
    None, # 0
    (1, TType.I64, 'version', None, None, ), # 1
    (2, TType.STRING, 'name', None, None, ), # 2
    (3, TType.STRING, 'owner', None, None, ), # 3
    (4, TType.I64, 'expirationTime', None, None, ), # 4
    (5, TType.DOUBLE, 'priority', None, 0, ), # 5
    (6, TType.STRING, 'data', None, None, ), # 6
  )

  def __init__(self, version=None, name=None, owner=None, expirationTime=None, priority=thrift_spec[5][4], data=None,):
    self.version = version
    self.name = name
    self.owner = owner
    self.expirationTime = expirationTime
    self.priority = priority
    self.data = data

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.I64:
          self.version = iprot.readI64();
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.name = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.STRING:
          self.owner = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.I64:
          self.expirationTime = iprot.readI64();
        else:
          iprot.skip(ftype)
      elif fid == 5:
        if ftype == TType.DOUBLE:
          self.priority = iprot.readDouble();
        else:
          iprot.skip(ftype)
      elif fid == 6:
        if ftype == TType.STRING:
          self.data = iprot.readString();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('Token')
    if self.version is not None:
      oprot.writeFieldBegin('version', TType.I64, 1)
      oprot.writeI64(self.version)
      oprot.writeFieldEnd()
    if self.name is not None:
      oprot.writeFieldBegin('name', TType.STRING, 2)
      oprot.writeString(self.name)
      oprot.writeFieldEnd()
    if self.owner is not None:
      oprot.writeFieldBegin('owner', TType.STRING, 3)
      oprot.writeString(self.owner)
      oprot.writeFieldEnd()
    if self.expirationTime is not None:
      oprot.writeFieldBegin('expirationTime', TType.I64, 4)
      oprot.writeI64(self.expirationTime)
      oprot.writeFieldEnd()
    if self.priority is not None:
      oprot.writeFieldBegin('priority', TType.DOUBLE, 5)
      oprot.writeDouble(self.priority)
      oprot.writeFieldEnd()
    if self.data is not None:
      oprot.writeFieldBegin('data', TType.STRING, 6)
      oprot.writeString(self.data)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.version is None:
      raise TProtocol.TProtocolException(message='Required field version is unset!')
    if self.name is None:
      raise TProtocol.TProtocolException(message='Required field name is unset!')
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class TokenMasterException(TException):
  """
  Attributes:
   - errorCode
   - errorMessage
  """

  thrift_spec = (
    None, # 0
    (1, TType.I32, 'errorCode', None, None, ), # 1
    (2, TType.STRING, 'errorMessage', None, None, ), # 2
  )

  def __init__(self, errorCode=None, errorMessage=None,):
    self.errorCode = errorCode
    self.errorMessage = errorMessage

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.I32:
          self.errorCode = iprot.readI32();
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.errorMessage = iprot.readString();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('TokenMasterException')
    if self.errorCode is not None:
      oprot.writeFieldBegin('errorCode', TType.I32, 1)
      oprot.writeI32(self.errorCode)
      oprot.writeFieldEnd()
    if self.errorMessage is not None:
      oprot.writeFieldBegin('errorMessage', TType.STRING, 2)
      oprot.writeString(self.errorMessage)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __str__(self):
    return repr(self)

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class ArchiveRequest:
  """
  Attributes:
   - tokens
  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'tokens', (TType.STRUCT,(Token, Token.thrift_spec)), None, ), # 1
  )

  def __init__(self, tokens=None,):
    self.tokens = tokens

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.tokens = []
          (_etype3, _size0) = iprot.readListBegin()
          for _i4 in xrange(_size0):
            _elem5 = Token()
            _elem5.read(iprot)
            self.tokens.append(_elem5)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('ArchiveRequest')
    if self.tokens is not None:
      oprot.writeFieldBegin('tokens', TType.LIST, 1)
      oprot.writeListBegin(TType.STRUCT, len(self.tokens))
      for iter6 in self.tokens:
        iter6.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class GroupRequest:
  """
  Attributes:
   - namePrefix
   - groupSuffix
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'namePrefix', None, None, ), # 1
    (2, TType.STRING, 'groupSuffix', None, None, ), # 2
  )

  def __init__(self, namePrefix=None, groupSuffix=None,):
    self.namePrefix = namePrefix
    self.groupSuffix = groupSuffix

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.namePrefix = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.groupSuffix = iprot.readString();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('GroupRequest')
    if self.namePrefix is not None:
      oprot.writeFieldBegin('namePrefix', TType.STRING, 1)
      oprot.writeString(self.namePrefix)
      oprot.writeFieldEnd()
    if self.groupSuffix is not None:
      oprot.writeFieldBegin('groupSuffix', TType.STRING, 2)
      oprot.writeString(self.groupSuffix)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class GroupResponse:
  """
  Attributes:
   - counts
  """

  thrift_spec = (
    None, # 0
    (1, TType.MAP, 'counts', (TType.STRING,None,TType.I32,None), None, ), # 1
  )

  def __init__(self, counts=None,):
    self.counts = counts

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.MAP:
          self.counts = {}
          (_ktype8, _vtype9, _size7 ) = iprot.readMapBegin() 
          for _i11 in xrange(_size7):
            _key12 = iprot.readString();
            _val13 = iprot.readI32();
            self.counts[_key12] = _val13
          iprot.readMapEnd()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('GroupResponse')
    if self.counts is not None:
      oprot.writeFieldBegin('counts', TType.MAP, 1)
      oprot.writeMapBegin(TType.STRING, TType.I32, len(self.counts))
      for kiter14,viter15 in self.counts.items():
        oprot.writeString(kiter14)
        oprot.writeI32(viter15)
      oprot.writeMapEnd()
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class ModifyRequest:
  """
  Attributes:
   - updates
   - deletes
  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'updates', (TType.STRUCT,(Token, Token.thrift_spec)), None, ), # 1
    (2, TType.LIST, 'deletes', (TType.STRUCT,(Token, Token.thrift_spec)), None, ), # 2
  )

  def __init__(self, updates=None, deletes=None,):
    self.updates = updates
    self.deletes = deletes

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.updates = []
          (_etype19, _size16) = iprot.readListBegin()
          for _i20 in xrange(_size16):
            _elem21 = Token()
            _elem21.read(iprot)
            self.updates.append(_elem21)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.LIST:
          self.deletes = []
          (_etype25, _size22) = iprot.readListBegin()
          for _i26 in xrange(_size22):
            _elem27 = Token()
            _elem27.read(iprot)
            self.deletes.append(_elem27)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('ModifyRequest')
    if self.updates is not None:
      oprot.writeFieldBegin('updates', TType.LIST, 1)
      oprot.writeListBegin(TType.STRUCT, len(self.updates))
      for iter28 in self.updates:
        iter28.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.deletes is not None:
      oprot.writeFieldBegin('deletes', TType.LIST, 2)
      oprot.writeListBegin(TType.STRUCT, len(self.deletes))
      for iter29 in self.deletes:
        iter29.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class ModifyResponse:
  """
  Attributes:
   - updates
  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'updates', (TType.STRUCT,(Token, Token.thrift_spec)), None, ), # 1
  )

  def __init__(self, updates=None,):
    self.updates = updates

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.updates = []
          (_etype33, _size30) = iprot.readListBegin()
          for _i34 in xrange(_size30):
            _elem35 = Token()
            _elem35.read(iprot)
            self.updates.append(_elem35)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('ModifyResponse')
    if self.updates is not None:
      oprot.writeFieldBegin('updates', TType.LIST, 1)
      oprot.writeListBegin(TType.STRUCT, len(self.updates))
      for iter36 in self.updates:
        iter36.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class Query:
  """
  Attributes:
   - namePrefix
   - maxTokens
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'namePrefix', None, None, ), # 1
    (2, TType.I32, 'maxTokens', None, None, ), # 2
  )

  def __init__(self, namePrefix=None, maxTokens=None,):
    self.namePrefix = namePrefix
    self.maxTokens = maxTokens

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.namePrefix = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I32:
          self.maxTokens = iprot.readI32();
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('Query')
    if self.namePrefix is not None:
      oprot.writeFieldBegin('namePrefix', TType.STRING, 1)
      oprot.writeString(self.namePrefix)
      oprot.writeFieldEnd()
    if self.maxTokens is not None:
      oprot.writeFieldBegin('maxTokens', TType.I32, 2)
      oprot.writeI32(self.maxTokens)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class QueryRequest:
  """
  Attributes:
   - queries
  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'queries', (TType.STRUCT,(Query, Query.thrift_spec)), None, ), # 1
  )

  def __init__(self, queries=None,):
    self.queries = queries

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.queries = []
          (_etype40, _size37) = iprot.readListBegin()
          for _i41 in xrange(_size37):
            _elem42 = Query()
            _elem42.read(iprot)
            self.queries.append(_elem42)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('QueryRequest')
    if self.queries is not None:
      oprot.writeFieldBegin('queries', TType.LIST, 1)
      oprot.writeListBegin(TType.STRUCT, len(self.queries))
      for iter43 in self.queries:
        iter43.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class QueryResponse:
  """
  Attributes:
   - tokens
  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'tokens', (TType.LIST,(TType.STRUCT,(Token, Token.thrift_spec))), None, ), # 1
  )

  def __init__(self, tokens=None,):
    self.tokens = tokens

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.tokens = []
          (_etype47, _size44) = iprot.readListBegin()
          for _i48 in xrange(_size44):
            _elem49 = []
            (_etype53, _size50) = iprot.readListBegin()
            for _i54 in xrange(_size50):
              _elem55 = Token()
              _elem55.read(iprot)
              _elem49.append(_elem55)
            iprot.readListEnd()
            self.tokens.append(_elem49)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('QueryResponse')
    if self.tokens is not None:
      oprot.writeFieldBegin('tokens', TType.LIST, 1)
      oprot.writeListBegin(TType.LIST, len(self.tokens))
      for iter56 in self.tokens:
        oprot.writeListBegin(TType.STRUCT, len(iter56))
        for iter57 in iter56:
          iter57.write(oprot)
        oprot.writeListEnd()
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class QueryAndOwnRequest:
  """
  Attributes:
   - owner
   - expirationTime
   - query
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'owner', None, None, ), # 1
    (2, TType.I64, 'expirationTime', None, None, ), # 2
    (3, TType.STRUCT, 'query', (Query, Query.thrift_spec), None, ), # 3
  )

  def __init__(self, owner=None, expirationTime=None, query=None,):
    self.owner = owner
    self.expirationTime = expirationTime
    self.query = query

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.owner = iprot.readString();
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I64:
          self.expirationTime = iprot.readI64();
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.STRUCT:
          self.query = Query()
          self.query.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('QueryAndOwnRequest')
    if self.owner is not None:
      oprot.writeFieldBegin('owner', TType.STRING, 1)
      oprot.writeString(self.owner)
      oprot.writeFieldEnd()
    if self.expirationTime is not None:
      oprot.writeFieldBegin('expirationTime', TType.I64, 2)
      oprot.writeI64(self.expirationTime)
      oprot.writeFieldEnd()
    if self.query is not None:
      oprot.writeFieldBegin('query', TType.STRUCT, 3)
      self.query.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class QueryAndOwnResponse:
  """
  Attributes:
   - tokens
  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'tokens', (TType.STRUCT,(Token, Token.thrift_spec)), None, ), # 1
  )

  def __init__(self, tokens=None,):
    self.tokens = tokens

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.tokens = []
          (_etype61, _size58) = iprot.readListBegin()
          for _i62 in xrange(_size58):
            _elem63 = Token()
            _elem63.read(iprot)
            self.tokens.append(_elem63)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('QueryAndOwnResponse')
    if self.tokens is not None:
      oprot.writeFieldBegin('tokens', TType.LIST, 1)
      oprot.writeListBegin(TType.STRUCT, len(self.tokens))
      for iter64 in self.tokens:
        iter64.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import mock
from pinball_ext.common.hadoop_utils import HadoopHostConfig
from pinball_ext.executor import emr_executor
from pinball_ext.executor.common import Platform


__author__ = 'Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


hadoop_node = HadoopHostConfig()
hadoop_node.USER_NAME = 'test_user'
hadoop_node.HOST_NAME = 'test_host_name'
hadoop_node.SSH_PORT = 22
hadoop_node.SSH_KEY_FILE = '~/test.key'
hadoop_node.HOST_NAME = '/usr/hadoop/'


class EmrExecutorTest(unittest.TestCase):
    @mock.patch('pinball_ext.common.hadoop_utils.run_and_check_command_in_hadoop')
    @mock.patch('pinball_ext.executor.emr_executor.EMRExecutor._get_raw_query_result',
                mock.Mock(return_value=[]))
    @mock.patch('pinball_ext.executor.emr_executor.EMRExecutor._mk_tmp_dir',
                mock.Mock(return_value='test_job_dir'))
    def test_hadoop_job(self, run_command):
        user_jar_dirs = ['/dir1/jar1', '/dir2/jar2']
        user_app_jar = '/dir/jar.jar'
        user_archive = '/dir/file.archive'

        executor = emr_executor.EMRExecutor(
            executor_config={
                'USER_LIBJAR_DIRS': ','.join(user_jar_dirs),
                'USER_APPJAR_PATH': user_app_jar,
                'USER_ARCHIVE_PATH': user_archive,
                'USER': 'test_user',

                'HADOOP_HOST_USER': hadoop_node.USER_NAME,
                'HADOOP_HOST_NAME': hadoop_node.HOST_NAME,
                'HADOOP_HOST_SSH_PORT': hadoop_node.SSH_PORT,
                'HADOOP_HOST_SSH_KEY_FILE': hadoop_node.SSH_KEY_FILE,
                'HADOOP_HOST_HOME': hadoop_node.REMOTE_HADOOP_HOME,
            }
        )
        self.assertEqual(executor.config.PLATFORM, Platform.EMR)

        executor.run_hadoop_job(
            'test_job_class',
            jobconf_args={
                'pinball.key1': 'value1',
                'pinball.key2': 'value2',
            },
            extra_args=[
                '-Dpinball.key3=value3'
            ])

        self.assertEqual(run_command.call_count, 1)
        self.assertEquals(len(run_command.call_args_list[0][0]), 2)
        self.assertEquals(run_command.call_args_list[0][0][0], hadoop_node)
        self.assertEquals(run_command.call_args_list[0][0][1],
                          "set -o pipefail; "
                          "HADOOP_CLASSPATH=/home/hadoop/test_user//dir1/jar1/*:/home/hadoop/test_user//dir2/jar2/* "
                          "hadoop jar /home/hadoop/test_user//dir/jar.jar test_job_class "
                          "-libjars `echo /home/hadoop/test_user//dir1/jar1/*.jar /home/hadoop/test_user//dir2/jar2/*.jar | tr ' ' ','` "
                          "-Dmapred.job.name=test_user:AdHocCommand "
                          "-Dpinball.key2=value2 "
                          "-Dpinball.key1=value1 "
                          "-Dpinball.key3=value3 "
                          "2>&1 > test_job_dir/out.csv | tee test_job_dir/out.err")
        self.assertEquals(run_command.call_args_list[0][1].keys(), ['log_line_processor'])


    @mock.patch('pinball_ext.common.hadoop_utils.put_string_to_hadoop')
    @mock.patch('pinball_ext.common.hadoop_utils.run_and_check_command_in_hadoop')
    @mock.patch('pinball_ext.executor.emr_executor.EMRExecutor._get_raw_query_result',
                mock.Mock(return_value=[]))
    @mock.patch('pinball_ext.executor.emr_executor.EMRExecutor._mk_tmp_dir',
                mock.Mock(return_value='test_query_dir'))
    def test_hive_job(self, run_command, put_string):
        executor = emr_executor.EMRExecutor(
            executor_config={
                'USER': 'test_user',
                'HADOOP_HOST_USER': hadoop_node.USER_NAME,
                'HADOOP_HOST_NAME': hadoop_node.HOST_NAME,
                'HADOOP_HOST_SSH_PORT': hadoop_node.SSH_PORT,
                'HADOOP_HOST_SSH_KEY_FILE': hadoop_node.SSH_KEY_FILE,
                'HADOOP_HOST_HOME': hadoop_node.REMOTE_HADOOP_HOME,
            }
        )

        q_str = 'SHOW TABLES;'
        executor.run_hive_query(q_str)

        self.assertEquals(put_string.call_count, 1)
        self.assertEquals(len(put_string.call_args_list[0][0]), 3)
        self.assertEquals(put_string.call_args_list[0][0][0], hadoop_node)
        self.assertEquals(put_string.call_args_list[0][0][1], 'set mapred.job.name=test_user:AdHocCommand;\nSHOW TABLES;')
        self.assertEquals(put_string.call_args_list[0][0][2], 'test_query_dir/query.ql')

        self.assertEquals(run_command.call_count, 1)
        self.assertEquals(len(run_command.call_args_list[0][0]), 2)
        self.assertEquals(run_command.call_args_list[0][0][0], hadoop_node)
        self.assertEquals(run_command.call_args_list[0][0][1],
                          'set -o pipefail; '
                          '/home/hadoop/hive/bin/hive -f test_query_dir/query.ql 2>&1 > test_query_dir/out.csv '
                          '| tee test_query_dir/out.err')
        self.assertEquals(run_command.call_args_list[0][1].keys(), ['log_line_processor'])



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parser utilities shared across modules."""
import calendar
import datetime
import pytz

from pinball.parser.config_parser import PARSER_CALLER_KEY
from pinball.workflow.utils import load_path


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


def recurrence_str_to_sec(recurrence_str):
    """Convert recurrence string to seconds value.

    Args:
        recurrence_str: The execution recurrence formatted as a numeric value
            and interval unit descriptor, e.b., 1d for a daily recurrence.
    Returns:
        Recurrence in seconds or None if input is misformatted.
    """
    if not recurrence_str or len(recurrence_str) < 2:
        return None
    value = int(recurrence_str[:-1])
    assert value > 0
    unit = recurrence_str[-1]
    if unit == 'w':
        return 7 * 24 * 60 * 60 * value
    elif unit == 'd':
        return 24 * 60 * 60 * value
    elif unit == 'H':
        return 60 * 60 * value
    elif unit == 'M':
        return 60 * value
    else:
        return None


def schedule_to_timestamp(execution_time, start_date=None):
    """Convert schedule specification to timestamp.

    Args:
        run_time: the execution time formatted as %H.%M.%S.%{millisecond}
        start_date: The first execution date formatted as %Y-%m-%d. If not
            present, we use the current date as the start date.
    Returns:
        Datetime object extracted from the schedule.
    """
    if start_date:
        date_arg = datetime.datetime.strptime(start_date, '%Y-%m-%d')
    else:
        date_arg = datetime.datetime.utcnow()
    time_arg = datetime.datetime.strptime(execution_time[:5], '%H.%M')
    parsed_datetime = datetime.datetime(date_arg.year,
                                        date_arg.month,
                                        date_arg.day,
                                        time_arg.hour,
                                        time_arg.minute,
                                        0,  # second
                                        0,  # microsecond
                                        pytz.utc)

    return int(calendar.timegm(parsed_datetime.timetuple()))


def annotate_parser_caller(parser_params, parser_caller):
    if parser_params:
        return dict({PARSER_CALLER_KEY: parser_caller}, **parser_params)
    else:
        return {PARSER_CALLER_KEY: parser_caller}


def load_parser_with_caller(parser_name, parser_params, parser_caller):
    return load_path(parser_name)(annotate_parser_caller(parser_params, parser_caller))



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generator of SVG diagram representing a given workflow instance."""
# pydot invokes 'dot' binary from the graphviz library.  This binary has to be
# accessible through $PATH.
import pydot

from pinball.ui.data import Status
from pinball.ui.utils import get_workflow_jobs_from_parser


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class WorkflowGraph(object):
    def __init__(self, jobs_data, instance_data):
        self._jobs_data = jobs_data
        self._instance_data = instance_data
        self._graph = pydot.Dot(graph_type='graph')
        # Graph id should be in sync with the value defined in
        # templates/jobs.html
        self._graph.set_id('workflow_graph')
        self._graph.set_orientation('portrait')
        self._build()

    @staticmethod
    def from_parser(workflow):
        data_jobs = get_workflow_jobs_from_parser(workflow)
        return WorkflowGraph(data_jobs, None)

    def _build(self):
        for job in self._jobs_data:
            node = pydot.Node(job.job)
            self._graph.add_node(node)
            node.set_id(job.job)
            node.set_fillcolor(Status.COLORS[job.status])
            node.add_style('filled')
            if job.is_condition:
                node.set_shape('octagon')
                if (job.status == Status.FAILURE and
                        self._instance_data and
                        self._instance_data.status == Status.RUNNING):
                    node.set_fillcolor(Status.COLORS[Status.CONDITION_PENDING])
            for output in job.outputs:
                edge = pydot.Edge(job.job, output)
                self._graph.add_edge(edge)
                edge.set_id('%s/%s' % (job.job, output))

    def get_svg(self):
        return self._graph.create_svg()



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for the workflow archiver."""
import mock
import time
import unittest

from pinball.workflow.archiver import Archiver
from pinball.master.thrift_lib.ttypes import ArchiveRequest
from pinball.master.thrift_lib.ttypes import Token
from pinball.master.thrift_lib.ttypes import Query
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.master.thrift_lib.ttypes import QueryResponse


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class ArchiverTestCase(unittest.TestCase):
    def setUp(self):
        self._client = mock.Mock()
        self._archiver = Archiver(self._client, 'some_workflow', '123')
        self._job_token = Token(
            version=12345,
            name='/workflow/some_workflow/123/waiting/some_job')
        self._abort_token = Token(
            version=123456,
            name='/workflow/some_workflow/123/__SIGNAL__/ABORT')

    def _prepare_get_instance_tokens(self, response_tokens):
        query_response = QueryResponse([response_tokens])
        self._client.query.return_value = query_response

    def _verify_get_instance_tokens(self):
        query = Query(namePrefix='/workflow/some_workflow/123/')
        query_request = QueryRequest(queries=[query])
        self._client.query.assert_called_once_with(query_request)

    def _verify_archive_tokens(self, request_tokens):
        archive_request = ArchiveRequest(tokens=request_tokens)
        self._client.archive.assert_called_once_with(archive_request)

    def test_archive_if_expired_non_existent(self):
        self._prepare_get_instance_tokens([])
        self.assertFalse(self._archiver.archive_if_expired(10))
        self._verify_get_instance_tokens()

    def test_archive_not_expired(self):
        self._prepare_get_instance_tokens([])
        self.assertFalse(self._archiver.archive_if_expired(time.time() + 1000))
        self.assertEqual(0, self._client.query.call_count)

    def test_archive_expired(self):
        self._prepare_get_instance_tokens([self._job_token])
        self.assertTrue(self._archiver.archive_if_expired(10))
        self._verify_get_instance_tokens()
        self._verify_archive_tokens([self._job_token])

    def test_archive_if_aborted_not_aborted(self):
        self._prepare_get_instance_tokens([self._job_token])
        self.assertFalse(self._archiver.archive_if_aborted())
        self._verify_get_instance_tokens()

    def test_archive_if_aborted_owned(self):
        self._job_token.owner = 'some_owner'
        self._job_token.expirationTime = time.time() + 1000
        self._prepare_get_instance_tokens([self._job_token, self._abort_token])
        self.assertFalse(self._archiver.archive_if_aborted())
        self._verify_get_instance_tokens()

    def test_archive_if_aborted(self):
        self._prepare_get_instance_tokens([self._job_token, self._abort_token])
        self.assertTrue(self._archiver.archive_if_aborted())
        self._verify_get_instance_tokens()
        self._verify_archive_tokens([self._job_token, self._abort_token])



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for utils."""
import collections
import datetime
import mock
import pytz
import unittest

from pinball.parser.config_parser import ParserCaller
from pinball.parser.config_parser import PARSER_CALLER_KEY
from pinball.parser.utils import annotate_parser_caller
from pinball.parser.utils import recurrence_str_to_sec
from pinball.parser.utils import schedule_to_timestamp


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class UtilsTestCase(unittest.TestCase):
    def test_recurrence_str_to_sec(self):
        self.assertEqual(10 * 60, recurrence_str_to_sec('10M'))
        self.assertEqual(10 * 60 * 60, recurrence_str_to_sec('10H'))
        self.assertEqual(10 * 24 * 60 * 60, recurrence_str_to_sec('10d'))
        self.assertEqual(10 * 7 * 24 * 60 * 60, recurrence_str_to_sec('10w'))
        self.assertIsNone(recurrence_str_to_sec('10Y'))

    def test_schedule_to_timestamp_with_start_date(self):
        # 1325376000 = 01 Jan 2012 00:00:00 UTC
        self.assertEqual(1325376000, schedule_to_timestamp('00.00.01.000',
                                                           '2012-01-01'))

    def test_schedule_to_timestamp_without_start_date(self):
        # utcnow cannot be monkey patched so we need to add a work around.
        datetime_now = datetime.datetime(2012, 1, 2, 0, 1, 0, 0, pytz.utc)
        Date = collections.namedtuple('date', 'year,month,day')
        date = Date(year=2012, month=1, day=2)
        with mock.patch('pinball.parser.utils.datetime.datetime'
                        ) as datetime_mock:
            datetime_mock.utcnow.return_value = date
            datetime_mock.return_value = datetime_now

            # 1325462460 = 02 Jan 2012 00:01:00 UTC
            self.assertEqual(1325462460, schedule_to_timestamp('00.00.01.000'))

    def test_annotate_parser_caller(self):
        org_param = {
            'key1': 'value1',
            'key2': 'value2',
            'key3': 3
        }
        new_param = annotate_parser_caller(org_param, ParserCaller.UI)

        self.assertEquals(len(org_param), 3)
        self.assertNotIn(PARSER_CALLER_KEY, org_param.keys())

        self.assertEquals(len(new_param), 4)
        self.assertIn(PARSER_CALLER_KEY, new_param.keys())

        org_param1 = None
        new_param1 = annotate_parser_caller(org_param1, ParserCaller.ANALYZER)
        self.assertIsNotNone(new_param1)
        self.assertEquals([PARSER_CALLER_KEY], new_param1.keys())



#
# Autogenerated by Thrift Compiler (0.9.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException
from ttypes import *
from thrift.Thrift import TProcessor
from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
  from thrift.protocol import fastbinary
except:
  fastbinary = None


class Iface:
  def archive(self, request):
    """
    Parameters:
     - request
    """
    pass

  def group(self, request):
    """
    Parameters:
     - request
    """
    pass

  def modify(self, request):
    """
    Parameters:
     - request
    """
    pass

  def query(self, request):
    """
    Parameters:
     - request
    """
    pass

  def query_and_own(self, request):
    """
    Parameters:
     - request
    """
    pass


class Client(Iface):
  def __init__(self, iprot, oprot=None):
    self._iprot = self._oprot = iprot
    if oprot is not None:
      self._oprot = oprot
    self._seqid = 0

  def archive(self, request):
    """
    Parameters:
     - request
    """
    self.send_archive(request)
    self.recv_archive()

  def send_archive(self, request):
    self._oprot.writeMessageBegin('archive', TMessageType.CALL, self._seqid)
    args = archive_args()
    args.request = request
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_archive(self, ):
    (fname, mtype, rseqid) = self._iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(self._iprot)
      self._iprot.readMessageEnd()
      raise x
    result = archive_result()
    result.read(self._iprot)
    self._iprot.readMessageEnd()
    if result.e is not None:
      raise result.e
    return

  def group(self, request):
    """
    Parameters:
     - request
    """
    self.send_group(request)
    return self.recv_group()

  def send_group(self, request):
    self._oprot.writeMessageBegin('group', TMessageType.CALL, self._seqid)
    args = group_args()
    args.request = request
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_group(self, ):
    (fname, mtype, rseqid) = self._iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(self._iprot)
      self._iprot.readMessageEnd()
      raise x
    result = group_result()
    result.read(self._iprot)
    self._iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    if result.e is not None:
      raise result.e
    raise TApplicationException(TApplicationException.MISSING_RESULT, "group failed: unknown result");

  def modify(self, request):
    """
    Parameters:
     - request
    """
    self.send_modify(request)
    return self.recv_modify()

  def send_modify(self, request):
    self._oprot.writeMessageBegin('modify', TMessageType.CALL, self._seqid)
    args = modify_args()
    args.request = request
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_modify(self, ):
    (fname, mtype, rseqid) = self._iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(self._iprot)
      self._iprot.readMessageEnd()
      raise x
    result = modify_result()
    result.read(self._iprot)
    self._iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    if result.e is not None:
      raise result.e
    raise TApplicationException(TApplicationException.MISSING_RESULT, "modify failed: unknown result");

  def query(self, request):
    """
    Parameters:
     - request
    """
    self.send_query(request)
    return self.recv_query()

  def send_query(self, request):
    self._oprot.writeMessageBegin('query', TMessageType.CALL, self._seqid)
    args = query_args()
    args.request = request
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_query(self, ):
    (fname, mtype, rseqid) = self._iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(self._iprot)
      self._iprot.readMessageEnd()
      raise x
    result = query_result()
    result.read(self._iprot)
    self._iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    if result.e is not None:
      raise result.e
    raise TApplicationException(TApplicationException.MISSING_RESULT, "query failed: unknown result");

  def query_and_own(self, request):
    """
    Parameters:
     - request
    """
    self.send_query_and_own(request)
    return self.recv_query_and_own()

  def send_query_and_own(self, request):
    self._oprot.writeMessageBegin('query_and_own', TMessageType.CALL, self._seqid)
    args = query_and_own_args()
    args.request = request
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_query_and_own(self, ):
    (fname, mtype, rseqid) = self._iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(self._iprot)
      self._iprot.readMessageEnd()
      raise x
    result = query_and_own_result()
    result.read(self._iprot)
    self._iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    if result.e is not None:
      raise result.e
    raise TApplicationException(TApplicationException.MISSING_RESULT, "query_and_own failed: unknown result");


class Processor(Iface, TProcessor):
  def __init__(self, handler):
    self._handler = handler
    self._processMap = {}
    self._processMap["archive"] = Processor.process_archive
    self._processMap["group"] = Processor.process_group
    self._processMap["modify"] = Processor.process_modify
    self._processMap["query"] = Processor.process_query
    self._processMap["query_and_own"] = Processor.process_query_and_own

  def process(self, iprot, oprot):
    (name, type, seqid) = iprot.readMessageBegin()
    if name not in self._processMap:
      iprot.skip(TType.STRUCT)
      iprot.readMessageEnd()
      x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
      oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
      x.write(oprot)
      oprot.writeMessageEnd()
      oprot.trans.flush()
      return
    else:
      self._processMap[name](self, seqid, iprot, oprot)
    return True

  def process_archive(self, seqid, iprot, oprot):
    args = archive_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = archive_result()
    try:
      self._handler.archive(args.request)
    except TokenMasterException as e:
      result.e = e
    oprot.writeMessageBegin("archive", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_group(self, seqid, iprot, oprot):
    args = group_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = group_result()
    try:
      result.success = self._handler.group(args.request)
    except TokenMasterException as e:
      result.e = e
    oprot.writeMessageBegin("group", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_modify(self, seqid, iprot, oprot):
    args = modify_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = modify_result()
    try:
      result.success = self._handler.modify(args.request)
    except TokenMasterException as e:
      result.e = e
    oprot.writeMessageBegin("modify", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_query(self, seqid, iprot, oprot):
    args = query_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = query_result()
    try:
      result.success = self._handler.query(args.request)
    except TokenMasterException as e:
      result.e = e
    oprot.writeMessageBegin("query", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_query_and_own(self, seqid, iprot, oprot):
    args = query_and_own_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = query_and_own_result()
    try:
      result.success = self._handler.query_and_own(args.request)
    except TokenMasterException as e:
      result.e = e
    oprot.writeMessageBegin("query_and_own", TMessageType.REPLY, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()


# HELPER FUNCTIONS AND STRUCTURES

class archive_args:
  """
  Attributes:
   - request
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'request', (ArchiveRequest, ArchiveRequest.thrift_spec), None, ), # 1
  )

  def __init__(self, request=None,):
    self.request = request

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.request = ArchiveRequest()
          self.request.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('archive_args')
    if self.request is not None:
      oprot.writeFieldBegin('request', TType.STRUCT, 1)
      self.request.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class archive_result:
  """
  Attributes:
   - e
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'e', (TokenMasterException, TokenMasterException.thrift_spec), None, ), # 1
  )

  def __init__(self, e=None,):
    self.e = e

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.e = TokenMasterException()
          self.e.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('archive_result')
    if self.e is not None:
      oprot.writeFieldBegin('e', TType.STRUCT, 1)
      self.e.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class group_args:
  """
  Attributes:
   - request
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'request', (GroupRequest, GroupRequest.thrift_spec), None, ), # 1
  )

  def __init__(self, request=None,):
    self.request = request

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.request = GroupRequest()
          self.request.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('group_args')
    if self.request is not None:
      oprot.writeFieldBegin('request', TType.STRUCT, 1)
      self.request.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class group_result:
  """
  Attributes:
   - success
   - e
  """

  thrift_spec = (
    (0, TType.STRUCT, 'success', (GroupResponse, GroupResponse.thrift_spec), None, ), # 0
    (1, TType.STRUCT, 'e', (TokenMasterException, TokenMasterException.thrift_spec), None, ), # 1
  )

  def __init__(self, success=None, e=None,):
    self.success = success
    self.e = e

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.STRUCT:
          self.success = GroupResponse()
          self.success.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 1:
        if ftype == TType.STRUCT:
          self.e = TokenMasterException()
          self.e.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('group_result')
    if self.success is not None:
      oprot.writeFieldBegin('success', TType.STRUCT, 0)
      self.success.write(oprot)
      oprot.writeFieldEnd()
    if self.e is not None:
      oprot.writeFieldBegin('e', TType.STRUCT, 1)
      self.e.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class modify_args:
  """
  Attributes:
   - request
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'request', (ModifyRequest, ModifyRequest.thrift_spec), None, ), # 1
  )

  def __init__(self, request=None,):
    self.request = request

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.request = ModifyRequest()
          self.request.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('modify_args')
    if self.request is not None:
      oprot.writeFieldBegin('request', TType.STRUCT, 1)
      self.request.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class modify_result:
  """
  Attributes:
   - success
   - e
  """

  thrift_spec = (
    (0, TType.STRUCT, 'success', (ModifyResponse, ModifyResponse.thrift_spec), None, ), # 0
    (1, TType.STRUCT, 'e', (TokenMasterException, TokenMasterException.thrift_spec), None, ), # 1
  )

  def __init__(self, success=None, e=None,):
    self.success = success
    self.e = e

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.STRUCT:
          self.success = ModifyResponse()
          self.success.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 1:
        if ftype == TType.STRUCT:
          self.e = TokenMasterException()
          self.e.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('modify_result')
    if self.success is not None:
      oprot.writeFieldBegin('success', TType.STRUCT, 0)
      self.success.write(oprot)
      oprot.writeFieldEnd()
    if self.e is not None:
      oprot.writeFieldBegin('e', TType.STRUCT, 1)
      self.e.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class query_args:
  """
  Attributes:
   - request
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'request', (QueryRequest, QueryRequest.thrift_spec), None, ), # 1
  )

  def __init__(self, request=None,):
    self.request = request

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.request = QueryRequest()
          self.request.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('query_args')
    if self.request is not None:
      oprot.writeFieldBegin('request', TType.STRUCT, 1)
      self.request.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class query_result:
  """
  Attributes:
   - success
   - e
  """

  thrift_spec = (
    (0, TType.STRUCT, 'success', (QueryResponse, QueryResponse.thrift_spec), None, ), # 0
    (1, TType.STRUCT, 'e', (TokenMasterException, TokenMasterException.thrift_spec), None, ), # 1
  )

  def __init__(self, success=None, e=None,):
    self.success = success
    self.e = e

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.STRUCT:
          self.success = QueryResponse()
          self.success.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 1:
        if ftype == TType.STRUCT:
          self.e = TokenMasterException()
          self.e.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('query_result')
    if self.success is not None:
      oprot.writeFieldBegin('success', TType.STRUCT, 0)
      self.success.write(oprot)
      oprot.writeFieldEnd()
    if self.e is not None:
      oprot.writeFieldBegin('e', TType.STRUCT, 1)
      self.e.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class query_and_own_args:
  """
  Attributes:
   - request
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'request', (QueryAndOwnRequest, QueryAndOwnRequest.thrift_spec), None, ), # 1
  )

  def __init__(self, request=None,):
    self.request = request

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.request = QueryAndOwnRequest()
          self.request.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('query_and_own_args')
    if self.request is not None:
      oprot.writeFieldBegin('request', TType.STRUCT, 1)
      self.request.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class query_and_own_result:
  """
  Attributes:
   - success
   - e
  """

  thrift_spec = (
    (0, TType.STRUCT, 'success', (QueryAndOwnResponse, QueryAndOwnResponse.thrift_spec), None, ), # 0
    (1, TType.STRUCT, 'e', (TokenMasterException, TokenMasterException.thrift_spec), None, ), # 1
  )

  def __init__(self, success=None, e=None,):
    self.success = success
    self.e = e

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.STRUCT:
          self.success = QueryAndOwnResponse()
          self.success.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 1:
        if ftype == TType.STRUCT:
          self.e = TokenMasterException()
          self.e.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('query_and_own_result')
    if self.success is not None:
      oprot.writeFieldBegin('success', TType.STRUCT, 0)
      self.success.write(oprot)
      oprot.writeFieldEnd()
    if self.e is not None:
      oprot.writeFieldBegin('e', TType.STRUCT, 1)
      self.e.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Format and send notification emails."""
import datetime
import smtplib

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pinball.config.pinball_config import PinballConfig
from pinball.config.utils import get_log
from pinball.config.utils import timestamp_to_str
from pinball.ui.data import Status


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = ['Pawel Garbacki', 'Mao Ye']
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.workflow.emailer')


class Emailer(object):
    """Send emails representing certain events."""
    def __init__(self, ui_host, ui_port):
        self._ui_host = ui_host
        self._ui_port = ui_port

    def _send_message(self, subject, to, text, html):
        """Send a message through local SMTP server.

        Args:
            subject: The subject of the email message.
            to: The list of recipient email addresses.
            text: The email body in text format.
            html: The email body in html format.
        """
        msg = MIMEMultipart('alternative')
        msg['Subject'] = subject
        msg['From'] = PinballConfig.DEFAULT_EMAIL
        msg['To'] = ', '.join(to)

        text_part = MIMEText(text, 'plain')
        html_part = MIMEText(html, 'html') if html else None

        # Attach parts into message container.  According to RFC 2046, the last
        # part of a multipart message, in this case the HTML message, is best
        # and preferred.
        msg.attach(text_part)
        if html_part:
            msg.attach(html_part)

        # Send the message via configurable SMTP server.
        if PinballConfig.SMTP_SSL:
            smtp = smtplib.SMTP_SSL(PinballConfig.SMTP_HOST, PinballConfig.SMTP_PORT)
        else:
            smtp = smtplib.SMTP(PinballConfig.SMTP_HOST, PinballConfig.SMTP_PORT)

        # Authorization: Login/Password
        if PinballConfig.SMTP_USER:
            smtp.login(PinballConfig.SMTP_USER, PinballConfig.SMTP_PASS)

        smtp.sendmail(msg['From'], to, msg.as_string())
        smtp.quit()
        LOG.info('Sent email to %s with subject "%s"', msg['To'], subject)

    def _get_instance_end_text(self, instance_data, jobs_data):
        """Format text version of the workflow instance completion message.

        Args:
            instance_data: The data of the completed workflow instance.
            jobs_data: The list of data describing jobs in the workflow
                instance.

        Returns:
            Text message describing the workflow instance.
        """
        TEXT_TEMPLATE = """\
Workflow %(workflow)s instance %(instance)s started at %(start_time)s \
finished after %(run_time)s at %(end_time)s with status %(status)s.

Details are available at  http://%(ui_host)s:%(ui_port)s/jobs/?\
workflow=%(workflow)s&instance=%(instance)s

Jobs:

Name | Last start | Last end | Run time | Status | Url
%(jobs)s
"""
        jobs = ''
        for job_data in jobs_data:
            jobs += '%s | ' % job_data.job
            if not job_data.last_start_time:
                jobs += '| | '
            else:
                jobs += '%s | ' % timestamp_to_str(job_data.last_start_time)
                if not job_data.last_end_time:
                    jobs += '| | '
                else:
                    jobs += '%s | ' % timestamp_to_str(job_data.last_end_time)
                    delta = int(job_data.last_end_time -
                                job_data.last_start_time)
                    jobs += '%s | ' % datetime.timedelta(seconds=delta)
            jobs += '%s | ' % Status.to_string(job_data.status)
            jobs += ('http://%s:%s/executions/?workflow=%s&instance=%s&'
                     'job=%s\n' % (self._ui_host,
                                   self._ui_port,
                                   job_data.workflow,
                                   job_data.instance,
                                   job_data.job))
        start_time = timestamp_to_str(instance_data.start_time)
        end_time = timestamp_to_str(instance_data.end_time)
        delta = int(instance_data.end_time - instance_data.start_time)
        run_time = datetime.timedelta(seconds=delta)
        status = Status.to_string(instance_data.status)
        return TEXT_TEMPLATE % {'workflow': instance_data.workflow,
                                'instance': instance_data.instance,
                                'start_time': start_time,
                                'end_time': end_time,
                                'run_time': run_time,
                                'status': status,
                                'ui_host': self._ui_host,
                                'ui_port': self._ui_port,
                                'jobs': jobs}

    def _get_instance_end_html(self, instance_data, jobs_data):
        """Format html version of the workflow instance completion message.

        Args:
            instance_data: The data of the completed workflow instance.
            jobs_data: The list of data describing jobs in the workflow
                instance.

        Returns:
            Html message describing the workflow instance.
        """
        HTML_TEMPLATE = """\
<html>
    <head></head>
    <body>
        <p>
            Workflow %(workflow)s instance %(instance)s started at
            %(start_time)s finished after %(run_time)s at %(end_time)s
            with status <span
style="background-color:%(instance_status_color)s;">%(status)s</span>.
            Click
            <a href="http://%(ui_host)s:%(ui_port)s/jobs/?\
workflow=%(workflow)s&instance=%(instance)s">here</a> for details.
        </p>
        <p>Jobs:<br/>
            <table style="border-collapse:collapse;">
                <tr>
                    <th style="border:1px dotted grey;">Name</th>
                    <th style="border:1px dotted grey;">Last start</th>
                    <th style="border:1px dotted grey;">Last end</th>
                    <th style="border:1px dotted grey;">Run time</th>
                    <th style="border:1px dotted grey;">Status</th>
                </tr>
                %(jobs)s
            </table>
        </p>
    </body>
</html>
"""
        jobs = ''
        for job_data in jobs_data:
            jobs += '<tr>'
            jobs += ('<td style="border:1px dotted grey;">'
                     '<a href="http://%s:%s/executions/?workflow=%s&'
                     'instance=%s&job=%s">%s</a></td>' % (self._ui_host,
                                                          self._ui_port,
                                                          job_data.workflow,
                                                          job_data.instance,
                                                          job_data.job,
                                                          job_data.job))
            if not job_data.last_start_time:
                jobs += ('<td style="border:1px dotted grey;"></td>'
                         '<td style="border:1px dotted grey;"></td>'
                         '<td style="border:1px dotted grey;"></td>')
            else:
                jobs += ('<td style="border:1px dotted grey;">%s</td>' %
                         timestamp_to_str(job_data.last_start_time))
                if not job_data.last_end_time:
                    jobs += ('<td style="border:1px dotted grey;"></td>'
                             '<td style="border:1px dotted grey;"></td>')
                else:
                    delta = int(job_data.last_end_time -
                                job_data.last_start_time)
                    jobs += ('<td style="border:1px dotted grey;">%s</td>'
                             '<td style="border:1px dotted grey;">%s</td>' % (
                                 timestamp_to_str(job_data.last_end_time),
                                 datetime.timedelta(seconds=delta)))
            jobs += ('<td style="border:1px dotted grey;background-color:%s;'
                     'text-align: center;">%s</td>\n' % (
                         Status.COLORS[job_data.status],
                         Status.to_string(job_data.status)))
            jobs += '</tr>'
        start_time = timestamp_to_str(instance_data.start_time)
        end_time = timestamp_to_str(instance_data.end_time)
        delta = int(instance_data.end_time - instance_data.start_time)
        run_time = datetime.timedelta(seconds=delta)
        instance_status_color = Status.COLORS[instance_data.status]
        status = Status.to_string(instance_data.status)
        return HTML_TEMPLATE % {'workflow': instance_data.workflow,
                                'instance': instance_data.instance,
                                'start_time': start_time,
                                'end_time': end_time,
                                'run_time': run_time,
                                'instance_status_color': instance_status_color,
                                'status': status,
                                'ui_host': self._ui_host,
                                'ui_port': self._ui_port,
                                'jobs': jobs}

    @staticmethod
    def _sort_jobs(jobs_data):
        """Sort jobs on the start time.

        Args:
            jobs_data: The list of jobs to sort.

        Returns:
            Sorted list of jobs.
        """
        def _start_time(jobs_data):
            return jobs_data.last_start_time
        return sorted(jobs_data, key=_start_time)

    def send_instance_end_message(self, to, instance_data, jobs_data):
        """Send a message describing workflow instance run.

        Args:
            to: The list of recipient email addresses.
            instance_data: The data of the completed workflow instance.
            jobs_data: The list of data describing jobs in the workflow
                instance.
        """
        jobs_data = Emailer._sort_jobs(jobs_data)
        text = self._get_instance_end_text(instance_data, jobs_data)
        html = self._get_instance_end_html(instance_data, jobs_data)
        status = Status.to_string(instance_data.status)
        subject = '%s for workflow %s' % (status, instance_data.workflow)
        self._send_message(subject, to, text, html)

    def _get_job_execution_params(self, job_execution_data):
        start_time = timestamp_to_str(job_execution_data.start_time)
        return {'workflow': job_execution_data.workflow,
                'instance': job_execution_data.instance,
                'job': job_execution_data.job,
                'execution': job_execution_data.execution,
                'info': job_execution_data.info,
                'start_time': start_time,
                'ui_host': self._ui_host,
                'ui_port': self._ui_port}

    def _get_job_execution_end_text(self, job_execution_data):
        """Format text version of the job execution completion message.

        Args:
            job_execution_data: The job execution data described in the
            message.
        Returns:
            Html message describing the job execution.
        """
        TEXT_TEMPLATE = """\
Job %(job)s execution in workflow %(workflow)s instance %(instance)s started \
at %(start_time)s finished after %(run_time)s at %(end_time)s with exit code \
%(exit_code)s.

Details are available at http://%(ui_host)s:%(ui_port)s/executions/?workflow=\
%(workflow)s&instance=%(instance)s&job=%(job)s

Execution details:

Workflow: %(workflow)s
Instance: %(instance)s
Job: %(job)s
Execution: %(execution)s
Info: %(info)s
Start time: %(start_time)s
End time: %(end_time)s
Run time: %(run_time)s
Logs: %(logs)s
Exit code: %(exit_code)s
"""
        params = self._get_job_execution_params(job_execution_data)
        end_time = timestamp_to_str(job_execution_data.end_time)
        params['end_time'] = end_time
        delta = int(job_execution_data.end_time -
                    job_execution_data.start_time)
        run_time = datetime.timedelta(seconds=delta)
        params['run_time'] = run_time
        logs = ', '.join(job_execution_data.logs.keys())
        params['logs'] = logs
        params['exit_code'] = job_execution_data.exit_code
        return TEXT_TEMPLATE % params

    def _get_job_execution_end_html(self, job_execution_data):
        """Format html version of the job execution completion message.

        Args:
            job_execution_data: The job execution data described in the
            message.
        Returns:
            Html message describing the job execution.
        """
        HTML_TEMPLATE = """\
<html>
    <head></head>
    <body>
        <p>
            Job %(job)s execution in workflow %(workflow)s instance
            %(instance)s started at %(start_time)s finished after %(run_time)s
            at %(end_time)s with exit code <span
            style="background-color:%(exit_code_color)s;">%(exit_code)s</span>.
            Click
            <a href="http://%(ui_host)s:%(ui_port)s/executions/?workflow=\
%(workflow)s&instance=%(instance)s&job=%(job)s">here</a> for details.
        </p>
        <p>Execution details:<br/>
            <table style="border-collapse:collapse;">
                <tr>
                    <td style="border:1px dotted grey;"><b>Workflow</b></td>
                    <td style="border:1px dotted grey;">
                        <a href="http://%(ui_host)s:%(ui_port)s/instances/?\
workflow=%(workflow)s">%(workflow)s</a>
                    </td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Instance</b></td>
                    <td style="border:1px dotted grey;">
                        <a href="http://%(ui_host)s:%(ui_port)s/jobs/?\
workflow=%(workflow)s&instance=%(instance)s">%(instance)s</a>
                    </td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Job</b></td>
                    <td style="border:1px dotted grey;">
                    <a href="http://%(ui_host)s:%(ui_port)s/executions/?\
workflow=%(workflow)s&instance=%(instance)s&job=%(job)s">%(job)s</a>
                    </td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Execution</b></td>
                    <td style="border:1px dotted grey;">%(execution)s</td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Info</b></td>
                    <td style="border:1px dotted grey;">%(info)s</td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Start time</b></td>
                    <td style="border:1px dotted grey;">%(start_time)s</td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>End time</b></td>
                    <td style="border:1px dotted grey;">%(end_time)s</td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Run time</b></td>
                    <td style="border:1px dotted grey;">%(run_time)s</td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Logs</b></td>
                    <td style="border:1px dotted grey;">%(logs)s</td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Exit code</b></td>
                    <td
style="border:1px dotted grey;background-color:%(exit_code_color)s;">
                        %(exit_code)s
                    </td>
                </tr>
            </table>
        </p>
    </body>
</html>
"""
        params = self._get_job_execution_params(job_execution_data)
        end_time = timestamp_to_str(job_execution_data.end_time)
        params['end_time'] = end_time
        delta = int(job_execution_data.end_time -
                    job_execution_data.start_time)
        run_time = datetime.timedelta(seconds=delta)
        params['run_time'] = run_time
        params['exit_code'] = job_execution_data.exit_code
        exit_code_color = (Status.COLORS[Status.SUCCESS] if
                           job_execution_data.exit_code == 0 else
                           Status.COLORS[Status.FAILURE])
        params['exit_code_color'] = exit_code_color
        logs = []
        for log_type in job_execution_data.logs:
            params['log_type'] = log_type
            logs.append('<a href="http://%(ui_host)s:%(ui_port)s/file/?'
                        'workflow=%(workflow)s&instance=%(instance)s&'
                        'job=%(job)s&execution=%(execution)s&'
                        'log_type=%(log_type)s">%(log_type)s</a>' % params)
        params['logs'] = '&nbsp;|&nbsp;'.join(logs)
        return HTML_TEMPLATE % params

    def send_job_execution_end_message(self, to, job_execution_data):
        """Send a message describing a job execution failure.

        Args:
            to: The list of recipient email addresses.
            job_execution_data: The job execution data described in the
                message.
        """
        text = self._get_job_execution_end_text(job_execution_data)
        html = self._get_job_execution_end_html(job_execution_data)

        subject = "Workflow %s's job %s finished with exit code %d" % (
            job_execution_data.workflow, job_execution_data.job,
            job_execution_data.exit_code)
        self._send_message(subject, to, text, html)

    def _get_job_timeout_warning_text(self, job_execution_data):
        """Format text version of the job timeout message.

        Args:
            job_execution_data: The job execution data described in the
            message.
        Returns:
            text message describing the job execution.
        """
        TEXT_TEMPLATE = """\
Job %(job)s execution in workflow %(workflow)s instance %(instance)s started \
at %(start_time)s reached timeout.

Details are available at http://%(ui_host)s:%(ui_port)s/executions/?workflow=\
%(workflow)s&instance=%(instance)s&job=%(job)s

Execution details:

Workflow: %(workflow)s
Instance: %(instance)s
Job: %(job)s
Execution: %(execution)s
Info: %(info)s
Start time: %(start_time)s
Logs: %(logs)s
"""
        params = self._get_job_execution_params(job_execution_data)
        logs = ', '.join(job_execution_data.logs.keys())
        params['logs'] = logs
        return TEXT_TEMPLATE % params

    def _get_job_timeout_warning_html(self, job_execution_data):
        """Format html version of the job timeout message.

        Args:
            job_execution_data: The job execution data described in the
            message.
        Returns:
            Html message describing the job execution.
        """
        HTML_TEMPLATE = """\
<html>
    <head></head>
    <body>
        <p>
            Job %(job)s execution in workflow %(workflow)s instance
            %(instance)s started at %(start_time)s reached timeout.
            Click
            <a href="http://%(ui_host)s:%(ui_port)s/executions/?workflow=\
%(workflow)s&instance=%(instance)s&job=%(job)s">here</a> for details.
        </p>
        <p>Execution details:<br/>
            <table style="border-collapse:collapse;">
                <tr>
                    <td style="border:1px dotted grey;"><b>Workflow</b></td>
                    <td style="border:1px dotted grey;">
                        <a href="http://%(ui_host)s:%(ui_port)s/instances/?\
workflow=%(workflow)s">%(workflow)s</a>
                    </td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Instance</b></td>
                    <td style="border:1px dotted grey;">
                        <a href="http://%(ui_host)s:%(ui_port)s/jobs/?\
workflow=%(workflow)s&instance=%(instance)s">%(instance)s</a>
                    </td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Job</b></td>
                    <td style="border:1px dotted grey;">
                    <a href="http://%(ui_host)s:%(ui_port)s/executions/?\
workflow=%(workflow)s&instance=%(instance)s&job=%(job)s">%(job)s</a>
                    </td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Execution</b></td>
                    <td style="border:1px dotted grey;">%(execution)s</td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Info</b></td>
                    <td style="border:1px dotted grey;">%(info)s</td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Start time</b></td>
                    <td style="border:1px dotted grey;">%(start_time)s</td>
                </tr>
                <tr>
                    <td style="border:1px dotted grey;"><b>Logs</b></td>
                    <td style="border:1px dotted grey;">%(logs)s</td>
                </tr>
            </table>
        </p>
    </body>
</html>
"""
        params = self._get_job_execution_params(job_execution_data)
        logs = []
        for log_type in job_execution_data.logs:
            params['log_type'] = log_type
            logs.append('<a href="http://%(ui_host)s:%(ui_port)s/file/?'
                        'workflow=%(workflow)s&instance=%(instance)s&'
                        'job=%(job)s&execution=%(execution)s&'
                        'log_type=%(log_type)s">%(log_type)s</a>' % params)
        params['logs'] = '&nbsp;|&nbsp;'.join(logs)
        return HTML_TEMPLATE % params

    def send_job_timeout_warning_message(self, to, job_execution_data):
        """Send a message warning about job execution exceeding timeout.

        Args:
            to: The list of recipient email addresses.
            job_execution_data: The job execution data described in the
                message.
        """
        text = self._get_job_timeout_warning_text(job_execution_data)
        html = self._get_job_timeout_warning_html(job_execution_data)

        subject = "Workflow %s's job %s exceeded timeout" % (
            job_execution_data.workflow, job_execution_data.job)
        try:
            self._send_message(subject, to, text, html)
        except:
            # Do not fail the job if sending out warning email failed.
            LOG.exception('failed to send warning email to %s for job %s'
                          % (to, job_execution_data.job))

    def send_too_many_running_instances_warning_message(self, to, workflow,
                                                        number_running_instances,
                                                        max_running_instances):
        """Send a message warning about number of running instances of a
        workflow exceeding the threshold.

        Args:
            to: Recipients of this email.
            workflow: Name of the workflow that the email is warning about.
            number_running_instances: Number of the current running instance for the
             given workflow.
            max_running_instances: Running instances of a workflow should lower
             than this number.
        """
        text = self._get_too_many_running_instances_warning_text(workflow,
                                                                 number_running_instances,
                                                                 max_running_instances)

        subject = "Too many (%s) instances running for workflow %s !" % (
            number_running_instances, workflow)

        try:
            self._send_message(subject, to, text, None)
        except:
            # Do not fail the job if sending out warning email failed.
            LOG.exception('failed to send too many instance running warning email to %s for workflow %s'
                          % (to, workflow))

    def _get_workflow_params(self, workflow, number_running_instances, max_running_instances):
        return {'ui_host': self._ui_host,
                'ui_port': self._ui_port,
                'running_instance_number': str(number_running_instances),
                'workflow': workflow,
                'max_running_instances': str(max_running_instances)}

    def _get_too_many_running_instances_warning_text(self,
                                                     workflow,
                                                     number_running_instances,
                                                     max_running_instances):
        """Format text version of too many running instance warning message.

        Args:
            workflow: Name of the workflow that the email is warning about.
            number_running_instances: Number of the current running instance for the
             given workflow.
            max_running_instances: Running instances of a workflow should lower than this number.

        Returns:
            Text message describing the workflow is running too many instances.
        """
        TEXT_TEMPLATE = """\
There are %(running_instance_number)s instance running at the same for workflow %(workflow)s,\
exceeding the threshold of %(max_running_instances)s!

Details are available at  http://%(ui_host)s:%(ui_port)s/instances/?workflow=%(workflow)s
"""
        params = self._get_workflow_params(workflow, number_running_instances,
                                           max_running_instances)

        return TEXT_TEMPLATE % params



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for utils."""
import pickle
import re
import unittest

from pinball.config.utils import get_unique_name
from pinball.config.utils import timestamp_to_str
from pinball.config.utils import token_to_str
from pinball.config.utils import str_to_timestamp
from pinball.master.thrift_lib.ttypes import Token


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class UtilsTestCase(unittest.TestCase):
    def test_get_unique_name(self):
        """Verify that unique name contains only allowed characters."""
        name = get_unique_name()
        self.assertIsNotNone(re.match(r'^\w+$', name))

    def test_token_to_string(self):
        """Test conversion of token to string."""
        token = Token(version=12345, name='/some_name', owner='some_owner',
                      expirationTime=10, priority=1.5,
                      data=pickle.dumps('some_data'))
        self.assertEqual('Token(version=12345, owner=some_owner, '
                         'expirationTime=1970-01-01 00:00:10 UTC, '
                         'priority=1.500000, name=/some_name, data=some_data)',
                         token_to_str(token))

    def test_timestamp_to_from_str(self):
        self.assertEqual('1970-01-01 00:00:10 UTC', timestamp_to_str(10))
        self.assertEqual(10, str_to_timestamp('1970-01-01 00:00:10 UTC'))



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from pinball_ext.job import basic_jobs


__author__ = 'Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class PythonTestJob(basic_jobs.PythonJob):
    def _execute(self):
        out_stdout = "%s/%s" % ('output', 'stdout')
        out_stderr = "%s/%s" % ('output', 'stderr')
        self._job_output = out_stdout
        self._job_stderr = out_stderr
        return


class CommandLineTestJob(basic_jobs.CommandLineJob):
    def _get_command(self):
        return 'echo "hello_world"'

    def _execute(self):
        super(CommandLineTestJob, self)._execute()
        self._job_stderr = 'stderr'
        return


class BasicJobTestCase(unittest.TestCase):
    def test_python_job_dryrun(self):
        job_py = PythonTestJob()
        job_py.runjob(dry_run=True)
        self.assertEquals(job_py._job_output, None)
        self.assertEquals(job_py._job_stderr, None)

    def test_python_job_normal(self):
        job_py = PythonTestJob()
        job_py.runjob()
        self.assertEquals(job_py._job_output, 'output/stdout')
        self.assertEquals(job_py._job_stderr, 'output/stderr')

    def test_command_line_dryrun(self):
        cl_job = CommandLineTestJob()
        cl_job.runjob(dry_run=True)
        self.assertEquals(cl_job._job_output, None)
        self.assertEquals(cl_job._job_stderr, None)

    def test_command_line_normal(self):
        cl_job = CommandLineTestJob()
        cl_job.runjob()
        self.assertEquals(cl_job._job_output, 'hello_world\n')
        self.assertEquals(cl_job._job_stderr, 'stderr')



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parser converting configurations to Pinball tokens."""
import abc


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'

PARSER_CALLER_KEY = 'caller'


class ParserCaller(object):
    ANALYZER = 'analyzer'
    SCHEDULE = 'schedule'
    UI = 'ui'
    WORKFLOW_UTIL = 'workflow_util'


class ConfigParser(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, params=None):
        """Create a config parser instance.

        Args:
            params: The free-form parser initialization parameters.
        """
        return

    @abc.abstractmethod
    def get_schedule_token(self, workflow):
        """Create schedule token for a given workflow.

        Args:
            workflow: The name of the workflow whose schedule token should be
                created.
        Returns:
            The workflow schedule token.
        """
        return

    @abc.abstractmethod
    def get_workflow_tokens(self, workflow):
        """Create tokens describing instance of a workflow.

        Args:
            workflow: The workflow whose tokens should be created.
        Returns:
            Job and event tokens defining a ready-to-run workflow instance.
        """
        return

    @abc.abstractmethod
    def get_workflow_names(self):
        """Get names of all workflows.

        Returns:
            Names of all workflows known to the parser.
        """
        return



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Overrun policy definition."""
from pinball.config.utils import PinballException


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class OverrunPolicy(object):
    """Overrun policy defines what to do if the previous run didn't finish.
    """
    SKIP, ABORT_RUNNING, DELAY, DELAY_UNTIL_SUCCESS, START_NEW = range(5)

    # Policy names and help strings.
    _POLICY_NAMES = {SKIP: ('SKIP', 'Skip execution if already running'),
                     ABORT_RUNNING: ('ABORT_RUNNING',
                                     'Abort the running instance before '
                                     'starting a new one'),
                     DELAY: ('DELAY', 'Delay the execution until the previous '
                             'one finishes'),
                     DELAY_UNTIL_SUCCESS: ('DELAY_UNTIL_SUCCESS', 'Delay the '
                                           'execution until the previous one '
                                           'succeeds'),
                     START_NEW: ('START_NEW', 'Start a new instance in '
                                 'parallel to currently running ones')}

    @staticmethod
    def to_string(policy):
        return OverrunPolicy._POLICY_NAMES[policy][0]

    @staticmethod
    def get_help(policy):
        return OverrunPolicy._POLICY_NAMES[policy][1]

    @staticmethod
    def from_string(policy_name):
        for policy, name in OverrunPolicy._POLICY_NAMES.items():
            if name[0] == policy_name:
                return policy
        raise PinballException('Unknown policy %s' % policy_name)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import unittest

from pinball.workflow.emailer import Emailer
from pinball.config.pinball_config import PinballConfig
from pinball.ui.data import JobData
from pinball.ui.data import JobExecutionData
from pinball.ui.data import Status
from pinball.ui.data import WorkflowInstanceData


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class EmailerTestCase(unittest.TestCase):
    @mock.patch('pinball.workflow.emailer.smtplib')
    def test_send_instance_end_message(self, smtplib_mock):
        emailer = Emailer('some_host', '8080')
        instance_data = WorkflowInstanceData(workflow='some_workflow',
                                             instance='123',
                                             status=Status.FAILURE,
                                             start_time=10,
                                             end_time=100)
        parent_job_data = JobData('some_workflow',
                                  '123',
                                  'parent',
                                  'ShellJob',
                                  False,
                                  'some_command',
                                  ['workflow_input'],
                                  ['parent_output'],
                                  ['some_email@pinterest.com'],
                                  1,
                                  2,
                                  100,
                                  200,
                                  1,
                                  Status.SUCCESS,
                                  10,
                                  50)

        child_job_data = JobData('some_workflow',
                                 '123',
                                 'child',
                                 'ShellJob',
                                 False,
                                 'some_command',
                                 ['parent_output'],
                                 [],
                                 ['some_email@pinterest.com'],
                                 1,
                                 2,
                                 100,
                                 200,
                                 2,
                                 Status.FAILURE,
                                 50,
                                 100)

        smtp = mock.Mock()
        smtplib_mock.SMTP.return_value = smtp

        sendmail = mock.Mock()
        smtp.sendmail = sendmail

        emailer.send_instance_end_message(['some_email@pinterest.com',
                                           'some_other_email@pinterest.com'],
                                          instance_data, [child_job_data,
                                                          parent_job_data])

        self.assertEqual(PinballConfig.DEFAULT_EMAIL, sendmail.call_args[0][0])
        self.assertEqual(
            ['some_email@pinterest.com', 'some_other_email@pinterest.com'],
            sendmail.call_args[0][1])
        msg = sendmail.call_args[0][2]
        self.assertTrue('FAILURE for workflow some_workflow' in msg)
        self.assertTrue('123' in msg)
        self.assertTrue('parent' in msg)
        self.assertTrue('child' in msg)
        self.assertTrue('SUCCESS' in msg)

        # Make sure that jobs are ordered on the end time.
        parent_index = msg.find('parent')
        child_index = msg.find('child')
        self.assertLess(parent_index, child_index)

    @mock.patch('pinball.workflow.emailer.smtplib')
    def test_send_job_execution_end_message(self, smtplib_mock):
        emailer = Emailer('some_host', '8080')
        execution_data = JobExecutionData(workflow='some_workflow',
                                          instance='123',
                                          job='some_job',
                                          execution=1,
                                          info='some_info',
                                          exit_code=1,
                                          start_time=10,
                                          end_time=100,
                                          logs={'stdout': '/some/out',
                                                'stderr': '/some/error'})

        smtp = mock.Mock()
        smtplib_mock.SMTP.return_value = smtp

        sendmail = mock.Mock()
        smtp.sendmail = sendmail

        emailer.send_job_execution_end_message(
            ['some_email@pinterest.com', 'some_other_email@pinterest.com'],
            execution_data)

        self.assertEqual(PinballConfig.DEFAULT_EMAIL, sendmail.call_args[0][0])
        self.assertEqual(
            ['some_email@pinterest.com', 'some_other_email@pinterest.com'],
            sendmail.call_args[0][1])
        msg = sendmail.call_args[0][2]
        self.assertTrue('Workflow some_workflow\'s job some_job finished with '
                        'exit code 1' in msg)
        self.assertTrue('123' in msg)
        self.assertTrue('some_job' in msg)
        self.assertTrue('stdout' in msg)
        self.assertTrue('stderr' in msg)

    @mock.patch('pinball.workflow.emailer.smtplib')
    def test_send_job_timeout_warning_message(self, smtplib_mock):
        emailer = Emailer('some_host', '8080')
        execution_data = JobExecutionData(workflow='some_workflow',
                                          instance='123',
                                          job='some_job',
                                          execution=1,
                                          info='some_info',
                                          start_time=10,
                                          logs={'stdout': '/some/out',
                                                'stderr': '/some/error'})

        smtp = mock.Mock()
        smtplib_mock.SMTP.return_value = smtp

        sendmail = mock.Mock()
        smtp.sendmail = sendmail

        emailer.send_job_timeout_warning_message(
            ['some_email@pinterest.com', 'some_other_email@pinterest.com'],
            execution_data)

        self.assertEqual(PinballConfig.DEFAULT_EMAIL, sendmail.call_args[0][0])
        self.assertEqual(
            ['some_email@pinterest.com', 'some_other_email@pinterest.com'],
            sendmail.call_args[0][1])
        msg = sendmail.call_args[0][2]
        self.assertTrue('Workflow some_workflow\'s job some_job exceeded '
                        'timeout' in msg)
        self.assertTrue('123' in msg)
        self.assertTrue('some_job' in msg)
        self.assertTrue('stdout' in msg)
        self.assertTrue('stderr' in msg)

    @mock.patch('pinball.workflow.emailer.smtplib')
    def test_send_too_many_running_instances_warning_message(self, smtplib_mock):
        emailer = Emailer('some_host', '8080')
        workflow = "some_workflow"
        number_running_instances = 3
        max_running_instances = 3

        smtp = mock.Mock()
        smtplib_mock.SMTP.return_value = smtp

        sendmail = mock.Mock()
        smtp.sendmail = sendmail

        emailer.send_too_many_running_instances_warning_message(
            ['some_email@pinterest.com', 'some_other_email@pinterest.com'],
            workflow,
            number_running_instances,
            max_running_instances)

        self.assertEqual(PinballConfig.DEFAULT_EMAIL, sendmail.call_args[0][0])
        self.assertEqual(
            ['some_email@pinterest.com', 'some_other_email@pinterest.com'],
            sendmail.call_args[0][1])
        msg = sendmail.call_args[0][2]
        self.assertTrue('3' in msg)
        self.assertTrue('some_workflow' in msg)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for buffered line reader"""

import unittest
import mock

from pinball.workflow import buffered_line_reader


__author__ = 'Jooseong Kim'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class BufferedLineReaderTestCase(unittest.TestCase):
    def setUp(self):
        file_mock = mock.Mock()
        self._file_mock = file_mock
        self._buffered_line_reader = \
            buffered_line_reader.BufferedLineReader(file_mock)

    @mock.patch('os.read')
    def test_one_line(self, os_read_mock):
        os_read_mock.return_value = 'one line'
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, [])
        self.assertFalse(self._buffered_line_reader.eof())

        os_read_mock.return_value = ''
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, ['one line'])
        self.assertTrue(self._buffered_line_reader.eof())

    @mock.patch('os.read')
    def test_one_line_continued(self, os_read_mock):
        os_read_mock.return_value = 'one line '
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, [])
        self.assertFalse(self._buffered_line_reader.eof())

        os_read_mock.return_value = 'continued'
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, [])
        self.assertFalse(self._buffered_line_reader.eof())

        os_read_mock.return_value = ''
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, ['one line continued'])
        self.assertTrue(self._buffered_line_reader.eof())

    @mock.patch('os.read')
    def test_one_line_terminated(self, os_read_mock):
        os_read_mock.return_value = 'one line terminated\n'
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, ['one line terminated\n'])
        self.assertFalse(self._buffered_line_reader.eof())

        os_read_mock.return_value = ''
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, [])
        self.assertTrue(self._buffered_line_reader.eof())

    @mock.patch('os.read')
    def test_one_line_continued_terminated(self, os_read_mock):
        os_read_mock.return_value = 'one line '
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, [])
        self.assertFalse(self._buffered_line_reader.eof())

        os_read_mock.return_value = 'continued\n'
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, ['one line continued\n'])
        self.assertFalse(self._buffered_line_reader.eof())

        os_read_mock.return_value = ''
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, [])
        self.assertTrue(self._buffered_line_reader.eof())

    @mock.patch('os.read')
    def test_multi_line(self, os_read_mock):
        os_read_mock.return_value = 'first line\nsecond line\nthird line'
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, ['first line\n', 'second line\n'])
        self.assertFalse(self._buffered_line_reader.eof())

        os_read_mock.return_value = ''
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, ['third line'])
        self.assertTrue(self._buffered_line_reader.eof())

    @mock.patch('os.read')
    def test_multi_line_continued_terminated(self, os_read_mock):
        os_read_mock.return_value = 'first '
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, [])
        self.assertFalse(self._buffered_line_reader.eof())

        os_read_mock.return_value = 'line\nsecond '
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, ['first line\n'])
        self.assertFalse(self._buffered_line_reader.eof())

        os_read_mock.return_value = 'line\nthird line\n'
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, ['second line\n', 'third line\n'])
        self.assertFalse(self._buffered_line_reader.eof())

        os_read_mock.return_value = ''
        lines = self._buffered_line_reader.readlines()
        self.assertEqual(lines, [])
        self.assertTrue(self._buffered_line_reader.eof())

    @mock.patch('os.read')
    def test_long_line(self, os_read_mock):
        """Tests if BufferedLineReader respects max_buffer_size"""
        a1009 = 'a' * 1009

        os_read_mock.return_value = a1009
        for i in range(15):
            lines = self._buffered_line_reader.readlines()
            os_read_mock.assert_called_with(self._file_mock.fileno(), 2048)
            self.assertEqual(lines, [])
            self.assertFalse(self._buffered_line_reader.eof())

        lines = self._buffered_line_reader.readlines()
        os_read_mock.assert_called_with(self._file_mock.fileno(), 1249)
        self.assertEqual(lines, [])
        self.assertFalse(self._buffered_line_reader.eof())

        os_read_mock.return_value = 'a' * 240
        lines = self._buffered_line_reader.readlines()
        os_read_mock.assert_called_with(self._file_mock.fileno(), 240)
        self.assertEqual(lines, ['a' * (1 << 14)])
        self.assertFalse(self._buffered_line_reader.eof())

        os_read_mock.return_value = a1009 + '\n'
        lines = self._buffered_line_reader.readlines()
        os_read_mock.assert_called_with(self._file_mock.fileno(), 2048)
        self.assertEqual(lines, [a1009 + '\n'])
        self.assertFalse(self._buffered_line_reader.eof())
